#ifndef EULER_KINETIC_SOLVER_HPP
#define EULER_KINETIC_SOLVER_HPP

#include <language/utils/FunctionSymbolId.hpp>
#include <scheme/DiscreteFunctionVariant.hpp>

std::tuple<std::shared_ptr<const DiscreteFunctionVariant>,
           std::shared_ptr<const DiscreteFunctionVariant>,
           std::shared_ptr<const DiscreteFunctionVariant>>
getEulerKineticWaves(const std::vector<double>& lambda_vector,
                     const std::shared_ptr<const DiscreteFunctionVariant>& rho,
                     const std::shared_ptr<const DiscreteFunctionVariant>& rho_u,
                     const std::shared_ptr<const DiscreteFunctionVariant>& rho_E,
                     const double& gamma);

std::tuple<std::shared_ptr<const DiscreteFunctionVariant>,
           std::shared_ptr<const DiscreteFunctionVariant>,
           std::shared_ptr<const DiscreteFunctionVariant>>
eulerKineticSolver(const double& dt,
                   const double& gamma,
                   const std::vector<double>& lambda_vector,
                   const double& eps,
                   const std::shared_ptr<const DiscreteFunctionVariant>& F_rho,
                   const std::shared_ptr<const DiscreteFunctionVariant>& F_rho_u,
                   const std::shared_ptr<const DiscreteFunctionVariant>& F_rho_E);

std::tuple<std::shared_ptr<const DiscreteFunctionVariant>, std::shared_ptr<const DiscreteFunctionVariant>>
getEulerKineticExactSolution(const double& t,
                             const std::shared_ptr<const DiscreteFunctionVariant>& rho_ex,
                             const std::shared_ptr<const DiscreteFunctionVariant>& u_ex);

#endif   // EULER_KINETIC_SOLVER_HPP
