#ifndef EULER_KINETIC_SOLVER_FIRST_ORDER_FV_HPP
#define EULER_KINETIC_SOLVER_FIRST_ORDER_FV_HPP

#include <language/utils/FunctionSymbolId.hpp>
#include <scheme/DiscreteFunctionVariant.hpp>

std::shared_ptr<const DiscreteFunctionVariant> refine(
  const std::shared_ptr<const DiscreteFunctionVariant>& old_u_old_mesh,
  const std::shared_ptr<const DiscreteFunctionVariant>& new_u_new_mesh);

std::shared_ptr<const DiscreteFunctionVariant> coarsen(
  const std::shared_ptr<const DiscreteFunctionVariant>& u_old_mesh,
  const std::shared_ptr<const DiscreteFunctionVariant>& new_u_new_mesh);

std::tuple<std::shared_ptr<const DiscreteFunctionVariant>,
           std::shared_ptr<const DiscreteFunctionVariant>,
           std::shared_ptr<const DiscreteFunctionVariant>>
getEulerKineticWavesFirstOrderFV(const std::vector<double>& lambda_vector,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& rho,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& rho_u,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& rho_E,
                                 const double& gamma);

std::tuple<std::shared_ptr<const DiscreteFunctionVariant>,
           std::shared_ptr<const DiscreteFunctionVariant>,
           std::shared_ptr<const DiscreteFunctionVariant>>
eulerKineticSolverFirstOrderFV(const double& dt,
                               const double& gamma,
                               const std::vector<double>& lambda_vector,
                               const double& eps,
                               const size_t& SpaceOrder,
                               const size_t& Limitation,
                               const bool& MoodLimitation,
                               const std::shared_ptr<const DiscreteFunctionVariant>& F_rho,
                               const std::shared_ptr<const DiscreteFunctionVariant>& F_rho_u,
                               const std::shared_ptr<const DiscreteFunctionVariant>& F_rho_E);

std::tuple<std::shared_ptr<const DiscreteFunctionVariant>,
           std::shared_ptr<const DiscreteFunctionVariant>,
           std::shared_ptr<const DiscreteFunctionVariant>,
           std::shared_ptr<const DiscreteFunctionVariant>,
           std::shared_ptr<const DiscreteFunctionVariant>>
getEulerKineticExactSolutionFirstOrderFV(const double& t,
                                         const double& gamma,
                                         const std::shared_ptr<const DiscreteFunctionVariant>& rho_ex,
                                         const std::shared_ptr<const DiscreteFunctionVariant>& u_ex,
                                         const std::shared_ptr<const DiscreteFunctionVariant>& p_ex,
                                         const std::shared_ptr<const DiscreteFunctionVariant>& rho_u_ex,
                                         const std::shared_ptr<const DiscreteFunctionVariant>& rho_E_ex);

#endif   // EULER_KINETIC_SOLVER_FIRST_ORDER_FV_HPP
