#ifndef EULER_KINETIC_SOLVER_MULTID_HPP
#define EULER_KINETIC_SOLVER_MULTID_HPP

#include <language/utils/FunctionSymbolId.hpp>
#include <scheme/DiscreteFunctionVariant.hpp>

class IBoundaryConditionDescriptor;

std::vector<TinyVector<2>> getLambdaVector2D(const std::shared_ptr<const MeshVariant>& mesh,
                                             const double& lambda,
                                             const size_t& L,
                                             const size_t& M);

std::vector<TinyVector<3>> getLambdaVector3D(const std::shared_ptr<const MeshVariant>& mesh, const double& lambda);

template <size_t Dimension>
std::tuple<std::shared_ptr<const DiscreteFunctionVariant>,
           std::shared_ptr<const DiscreteFunctionVariant>,
           std::shared_ptr<const DiscreteFunctionVariant>>
getEulerKineticWavesMultiD(const std::vector<TinyVector<Dimension>>& lambda_vector,
                           const std::shared_ptr<const DiscreteFunctionVariant>& rho,
                           const std::shared_ptr<const DiscreteFunctionVariant>& rho_u,
                           const std::shared_ptr<const DiscreteFunctionVariant>& rho_E,
                           const double& gamma);

template <size_t Dimension>
std::tuple<std::shared_ptr<const DiscreteFunctionVariant>,
           std::shared_ptr<const DiscreteFunctionVariant>,
           std::shared_ptr<const DiscreteFunctionVariant>>
eulerKineticSolverMultiD(const double& dt,
                         const double& gamma,
                         const std::vector<TinyVector<Dimension>>& lambda_vector,
                         const double& eps,
                         const size_t& SpaceOrder,
                         const size_t& scheme,
                         const std::shared_ptr<const DiscreteFunctionVariant>& F_rho,
                         const std::shared_ptr<const DiscreteFunctionVariant>& F_rho_u,
                         const std::shared_ptr<const DiscreteFunctionVariant>& F_rho_E,
                         const std::vector<std::shared_ptr<const IBoundaryConditionDescriptor>>& bc_descriptor_list);

#endif   // EULER_KINETIC_SOLVER_MULTID_HPP
