#include <scheme/EulerKineticSolverThirdOrderFD.hpp>

#include <language/utils/EvaluateAtPoints.hpp>
#include <mesh/Connectivity.hpp>
#include <mesh/Mesh.hpp>
#include <mesh/MeshVariant.hpp>
#include <scheme/DiscreteFunctionDescriptorP0Vector.hpp>
#include <scheme/DiscreteFunctionP0Vector.hpp>
#include <scheme/DiscreteFunctionUtils.hpp>
#include <scheme/IDiscreteFunctionDescriptor.hpp>

#include <tuple>

#include <analysis/GaussQuadratureDescriptor.hpp>
#include <analysis/QuadratureManager.hpp>

template <MeshConcept MeshType>
std::tuple<std::shared_ptr<const DiscreteFunctionVariant>, std::shared_ptr<const DiscreteFunctionVariant>>
getEulerKineticExactSolutionThirdOrderFD(const std::shared_ptr<const MeshType> p_mesh, const double t)
{
  DiscreteFunctionP0<double> rho_ex(p_mesh);
  DiscreteFunctionP0<double> u_ex(p_mesh);

  MeshData<MeshType>& mesh_data                                    = MeshDataManager::instance().getMeshData(*p_mesh);
  const CellValue<const TinyVector<MeshType::Dimension>>& x_vector = mesh_data.xj();

  // const QuadratureFormula<1>& l1 = QuadratureManager::instance().getLineFormula(GaussQuadratureDescriptor(5));

  const double pi = acos(-1);
  // EvaluateAtPoints<double(double)>::evaluateTo(density_function_id, x, rho0);

  for (CellId cell_id = 0; cell_id < p_mesh->numberOfCells(); ++cell_id) {
    double xj      = x_vector[cell_id][0];
    double a1      = -1.5;
    double b1      = 1.5;
    double c1      = 0.5 * (a1 + b1);
    double x1      = c1;
    double rho0_x1 = 1 + 0.5 * sin(2 * pi * x1);
    double rho0_a1 = 1 + 0.5 * sin(2 * pi * a1);
    double rho0_b1 = 1 + 0.5 * sin(2 * pi * b1);
    double rho0_c1 = 1 + 0.5 * sin(2 * pi * c1);

    double fx1 = xj + sqrt(3) * rho0_x1 * t - x1;
    double fa1 = xj + sqrt(3) * rho0_a1 * t - a1;
    double fb1 = xj + sqrt(3) * rho0_b1 * t - b1;
    double fc1 = xj + sqrt(3) * rho0_c1 * t - c1;

    while (std::abs(fx1) > 1E-15) {
      if (fa1 * fc1 < 0) {
        b1 = c1;
        c1 = 0.5 * (a1 + b1);
      } else if (fc1 * fb1 < 0) {
        a1 = c1;
        c1 = 0.5 * (a1 + b1);
      }
      x1 = c1;

      rho0_x1 = 1 + 0.5 * sin(2 * pi * x1);
      rho0_a1 = 1 + 0.5 * sin(2 * pi * a1);
      rho0_b1 = 1 + 0.5 * sin(2 * pi * b1);
      rho0_c1 = 1 + 0.5 * sin(2 * pi * c1);

      fx1 = xj + sqrt(3) * rho0_x1 * t - x1;
      fa1 = xj + sqrt(3) * rho0_a1 * t - a1;
      fb1 = xj + sqrt(3) * rho0_b1 * t - b1;
      fc1 = xj + sqrt(3) * rho0_c1 * t - c1;
    }

    double a2      = -1.5;
    double b2      = 1.5;
    double c2      = 0.5 * (a2 + b2);
    double x2      = c2;
    double rho0_x2 = 1 + 0.5 * sin(2 * pi * x2);
    double rho0_a2 = 1 + 0.5 * sin(2 * pi * a2);
    double rho0_b2 = 1 + 0.5 * sin(2 * pi * b2);
    double rho0_c2 = 1 + 0.5 * sin(2 * pi * c2);

    double fx2 = xj - sqrt(3) * rho0_x2 * t - x2;
    double fa2 = xj - sqrt(3) * rho0_a2 * t - a2;
    double fb2 = xj - sqrt(3) * rho0_b2 * t - b2;
    double fc2 = xj - sqrt(3) * rho0_c2 * t - c2;

    while (std::abs(fx2) > 1E-15) {
      if (fa2 * fc2 < 0) {
        b2 = c2;
        c2 = 0.5 * (a2 + b2);
      } else if (fc2 * fb2 < 0) {
        a2 = c2;
        c2 = 0.5 * (a2 + b2);
      }
      x2 = c2;

      rho0_x2 = 1 + 0.5 * sin(2 * pi * x2);
      rho0_a2 = 1 + 0.5 * sin(2 * pi * a2);
      rho0_b2 = 1 + 0.5 * sin(2 * pi * b2);
      rho0_c2 = 1 + 0.5 * sin(2 * pi * c2);

      fx2 = xj - sqrt(3) * rho0_x2 * t - x2;
      fa2 = xj - sqrt(3) * rho0_a2 * t - a2;
      fb2 = xj - sqrt(3) * rho0_b2 * t - b2;
      fc2 = xj - sqrt(3) * rho0_c2 * t - c2;
    }

    rho0_x1 = 1 + 0.5 * sin(2 * pi * x1);
    rho0_x2 = 1 + 0.5 * sin(2 * pi * x2);

    rho_ex[cell_id] = 0.5 * (rho0_x1 + rho0_x2);
    u_ex[cell_id]   = sqrt(3) * (rho_ex[cell_id] - rho0_x1);
  }

  return std::make_tuple(std::make_shared<DiscreteFunctionVariant>(rho_ex),
                         std::make_shared<DiscreteFunctionVariant>(u_ex));
}

std::tuple<std::shared_ptr<const DiscreteFunctionVariant>, std::shared_ptr<const DiscreteFunctionVariant>>
getEulerKineticExactSolutionThirdOrderFD(const double& t,
                                         const std::shared_ptr<const DiscreteFunctionVariant>& rho_ex,
                                         const std::shared_ptr<const DiscreteFunctionVariant>& u_ex)
{
  const std::shared_ptr<const MeshVariant> mesh_v = getCommonMesh({u_ex, rho_ex});
  if (mesh_v.use_count() == 0) {
    throw NormalError("rho_ex, and u_ex have to be defined on the same mesh");
  }

  return std::visit(
    [&](auto&& p_mesh)
      -> std::tuple<std::shared_ptr<const DiscreteFunctionVariant>, std::shared_ptr<const DiscreteFunctionVariant>> {
      using MeshType = mesh_type_t<decltype(p_mesh)>;
      if constexpr (is_polygonal_mesh_v<MeshType> and (MeshType::Dimension == 1)) {
        return getEulerKineticExactSolutionThirdOrderFD(p_mesh, t);
      } else {
        throw NotImplementedError("Invalid mesh type for Scalar Kinetic solver");
      }
    },
    mesh_v->variant());
}

template <MeshConcept MeshType>
std::tuple<std::shared_ptr<const DiscreteFunctionVariant>,
           std::shared_ptr<const DiscreteFunctionVariant>,
           std::shared_ptr<const DiscreteFunctionVariant>>
getEulerKineticWavesThirdOrderFD(const std::shared_ptr<const MeshType> p_mesh,
                                 const std::vector<double> lambda_vector,
                                 DiscreteFunctionP0<const double> rho,
                                 DiscreteFunctionP0<const double> rho_u,
                                 DiscreteFunctionP0<const double> rho_E,
                                 const double gamma)
{
  if (lambda_vector.size() < 2) {
    throw NormalError("lambda vector must be of dimension greater than 2");
  }

  DiscreteFunctionP0Vector<double> Fn_rho(p_mesh, lambda_vector.size());
  DiscreteFunctionP0Vector<double> Fn_rho_u(p_mesh, lambda_vector.size());
  DiscreteFunctionP0Vector<double> Fn_rho_E(p_mesh, lambda_vector.size());

  CellArray<double> vec_A{p_mesh->connectivity(), 3};
  CellValue<double> p{p_mesh->connectivity()};

  for (CellId cell_id = 0; cell_id < p_mesh->numberOfCells(); ++cell_id) {
    double rho_u_2    = (rho_u[cell_id] * rho_u[cell_id]) / rho[cell_id];
    p[cell_id]        = (gamma - 1) * (rho_E[cell_id] - 0.5 * rho_u_2);
    vec_A[cell_id][0] = rho_u[cell_id];
    vec_A[cell_id][1] = rho_u_2 + p[cell_id];
    vec_A[cell_id][2] = rho_u[cell_id] / rho[cell_id] * (rho_E[cell_id] * p[cell_id]);
  }

  const size_t nb_waves           = lambda_vector.size();
  double inv_L2norm_lambda_vector = 0;
  for (size_t i = 0; i < nb_waves; ++i) {
    inv_L2norm_lambda_vector += std::pow(lambda_vector[i], 2);
  }
  inv_L2norm_lambda_vector = 1. / inv_L2norm_lambda_vector;

  for (CellId cell_id = 0; cell_id < p_mesh->numberOfCells(); ++cell_id) {
    for (size_t i = 0; i < nb_waves; ++i) {
      Fn_rho[cell_id][i] =
        (1. / nb_waves) * rho[cell_id] + inv_L2norm_lambda_vector * vec_A[cell_id][0] * lambda_vector[i];
      Fn_rho_u[cell_id][i] =
        (1. / nb_waves) * rho_u[cell_id] + inv_L2norm_lambda_vector * vec_A[cell_id][1] * lambda_vector[i];
      Fn_rho_E[cell_id][i] =
        (1. / nb_waves) * rho_E[cell_id] + inv_L2norm_lambda_vector * vec_A[cell_id][2] * lambda_vector[i];
    }
  }

  return std::make_tuple(std::make_shared<DiscreteFunctionVariant>(Fn_rho),
                         std::make_shared<DiscreteFunctionVariant>(Fn_rho_u),
                         std::make_shared<DiscreteFunctionVariant>(Fn_rho_E));
}

std::tuple<std::shared_ptr<const DiscreteFunctionVariant>,
           std::shared_ptr<const DiscreteFunctionVariant>,
           std::shared_ptr<const DiscreteFunctionVariant>>
getEulerKineticWavesThirdOrderFD(const std::vector<double>& lambda_vector,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& rho,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& rho_u,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& rho_E,
                                 const double& gamma)
{
  const std::shared_ptr<const MeshVariant> mesh_v = getCommonMesh({rho, rho_u, rho_E});
  if (mesh_v.use_count() == 0) {
    throw NormalError("rho, rho_u and rho_E have to be defined on the same mesh");
  }

  return std::visit(
    [&](auto&& p_mesh)
      -> std::tuple<std::shared_ptr<const DiscreteFunctionVariant>, std::shared_ptr<const DiscreteFunctionVariant>,
                    std::shared_ptr<const DiscreteFunctionVariant>> {
      using MeshType = mesh_type_t<decltype(p_mesh)>;
      if constexpr (is_polygonal_mesh_v<MeshType> and (MeshType::Dimension == 1)) {
        auto rho_h   = rho->get<DiscreteFunctionP0<const double>>();
        auto rho_u_h = rho_u->get<DiscreteFunctionP0<const double>>();
        auto rho_E_h = rho_E->get<DiscreteFunctionP0<const double>>();

        return getEulerKineticWavesThirdOrderFD(p_mesh, lambda_vector, rho_h, rho_u_h, rho_E_h, gamma);
      } else {
        throw NotImplementedError("Invalid mesh type for Scalar Kinetic solver");
      }
    },
    mesh_v->variant());
}

template <MeshConcept MeshType>
class EulerKineticSolverThirdOrderFD
{
 private:
  constexpr static size_t Dimension = MeshType::Dimension;
  using Rd                          = TinyVector<Dimension>;

  const double m_dt;
  const std::vector<double>& m_lambda_vector;
  const DiscreteFunctionP0Vector<const double> m_Fn_rho;
  const DiscreteFunctionP0Vector<const double> m_Fn_rho_u;
  const DiscreteFunctionP0Vector<const double> m_Fn_rho_E;
  const double m_gamma;
  const std::shared_ptr<const MeshType> p_mesh;
  const MeshType& m_mesh;
  const CellValue<const double> m_dx_table          = MeshDataManager::instance().getMeshData(m_mesh).Vj();
  const CellValue<const TinyVector<Dimension>> m_xj = MeshDataManager::instance().getMeshData(m_mesh).xj();
  const NodeValue<const TinyVector<Dimension>> m_xr = m_mesh.xr();

 public:
  CellArray<double>
  getA(const CellValue<const double>& rho,
       const CellValue<const double>& rho_u,
       const CellValue<const double>& rho_E) const
  {
    CellArray<double> vec_A{m_mesh.connectivity(), 3};
    CellValue<double> p{m_mesh.connectivity()};
    for (CellId cell_id = 0; cell_id < m_mesh.numberOfCells(); ++cell_id) {
      double rho_u_2    = (rho_u[cell_id] * rho_u[cell_id]) / rho[cell_id];
      p[cell_id]        = (m_gamma - 1) * (rho_E[cell_id] - 0.5 * rho_u_2);
      vec_A[cell_id][0] = rho_u[cell_id];
      vec_A[cell_id][1] = rho_u_2 + p[cell_id];
      vec_A[cell_id][2] = rho_u[cell_id] / rho[cell_id] * (rho_E[cell_id] + p[cell_id]);
    }

    return vec_A;
  }

  CellArray<double>
  compute_M(const CellValue<const double>& u, const CellValue<const double> Au)
  {
    const size_t nb_waves = m_lambda_vector.size();
    CellArray<double> M{m_mesh.connectivity(), nb_waves};

    double inv_L2norm_lambda_vector = 0;
    for (size_t i = 0; i < nb_waves; ++i) {
      inv_L2norm_lambda_vector += pow(m_lambda_vector[i], 2);
    }
    inv_L2norm_lambda_vector = 1. / inv_L2norm_lambda_vector;

    for (CellId cell_id = 0; cell_id < m_mesh.numberOfCells(); ++cell_id) {
      for (size_t i = 0; i < nb_waves; ++i) {
        M[cell_id][i] = (1. / nb_waves) * u[cell_id] + inv_L2norm_lambda_vector * Au[cell_id] * m_lambda_vector[i];
      }
    }
    return M;
  }

  NodeArray<double>
  compute_Flux1(const DiscreteFunctionP0Vector<const double>& F, const NodeValue<const bool>& PF_node_is_valid)
  {
    const auto& cell_to_node_matrix = m_mesh.connectivity().cellToNodeMatrix();
    const size_t nb_waves           = m_lambda_vector.size();

    NodeArray<double> Flux_F{m_mesh.connectivity(), nb_waves};
    Flux_F.fill(0);

    for (CellId cell_id = 0; cell_id < m_mesh.numberOfCells(); ++cell_id) {
      for (size_t i = 0; i < nb_waves; ++i) {
        const size_t N = m_mesh.numberOfCells();

        const auto& cell_to_node   = cell_to_node_matrix[cell_id];
        const NodeId left_node_id  = cell_to_node[0];
        const NodeId right_node_id = cell_to_node[1];

        const CellId prev_cell_id = (cell_id + N - 1) % N;
        const CellId next_cell_id = (cell_id + 1) % N;

        if (m_lambda_vector[i] < 0) {
          if (PF_node_is_valid[left_node_id] == false) {
            Flux_F[left_node_id][i] = F[cell_id][i];
          } else if (PF_node_is_valid[right_node_id] == false) {
            Flux_F[right_node_id][i] = F[next_cell_id][i];
          }

        } else if (m_lambda_vector[i] > 0) {
          if (PF_node_is_valid[left_node_id] == false) {
            Flux_F[left_node_id][i] = F[prev_cell_id][i];
          } else if (PF_node_is_valid[right_node_id] == false) {
            Flux_F[right_node_id][i] = F[cell_id][i];
          }
        } else {
          Flux_F[right_node_id][i] = 0;
          Flux_F[left_node_id][i]  = 0;
        }
      }
    }
    return Flux_F;
  }

  DiscreteFunctionP0Vector<double>
  compute_delta1Fn(const DiscreteFunctionP0Vector<const double>& Fn)
  {
    const size_t nb_waves = m_lambda_vector.size();

    DiscreteFunctionP0Vector<double> deltaFn(p_mesh, nb_waves);

    for (CellId cell_id = 0; cell_id < m_mesh.numberOfCells(); ++cell_id) {
      for (size_t i = 0; i < m_lambda_vector.size(); ++i) {
        if (m_lambda_vector[i] < 0) {
          deltaFn[cell_id][i] = Fn[(cell_id + 1) % m_mesh.numberOfCells()][i] - Fn[cell_id][i];
        } else if (m_lambda_vector[i] > 0) {
          deltaFn[cell_id][i] = Fn[cell_id][i] - Fn[(cell_id + m_mesh.numberOfCells() - 1) % m_mesh.numberOfCells()][i];
        } else {
          deltaFn[cell_id][i] = 0;
        }
      }
    }

    return deltaFn;
  }

  DiscreteFunctionP0Vector<double>
  compute_delta3Fn(const DiscreteFunctionP0Vector<const double>& Fn)
  {
    const auto& cell_to_node_matrix = m_mesh.connectivity().cellToNodeMatrix();
    const size_t nb_waves           = m_lambda_vector.size();

    DiscreteFunctionP0Vector<double> deltaFn(p_mesh, nb_waves);

    for (CellId cell_id = 0; cell_id < m_mesh.numberOfCells(); ++cell_id) {
      for (size_t i = 0; i < nb_waves; ++i) {
        const NodeId first_node_id = 0;
        const NodeId last_node_id  = m_mesh.numberOfNodes() - 1;

        const double xmin = m_xr[first_node_id][0];
        const double xmax = m_xr[last_node_id][0];

        const size_t N = m_mesh.numberOfCells();

        const auto& cell_to_node   = cell_to_node_matrix[cell_id];
        const NodeId left_node_id  = cell_to_node[0];
        const NodeId right_node_id = cell_to_node[1];

        const CellId prev2_cell_id = (cell_id + N - 2) % N;
        const CellId prev_cell_id  = (cell_id + N - 1) % N;
        const CellId next_cell_id  = (cell_id + 1) % N;
        const CellId next2_cell_id = (cell_id + 2) % N;

        const double xjm2 = (-1 * (cell_id < 2) * (xmax - xmin)) + m_xj[prev2_cell_id][0];
        const double xjm1 = (-1 * (cell_id < 1) * (xmax - xmin)) + m_xj[prev_cell_id][0];
        const double xj   = m_xj[cell_id][0];
        const double xjp1 = ((cell_id + 1 >= N) * (xmax - xmin)) + m_xj[next_cell_id][0];
        const double xjp2 = ((cell_id + 2 >= N) * (xmax - xmin)) + m_xj[next2_cell_id][0];

        if (m_lambda_vector[i] > 0) {
          deltaFn[cell_id][i] =
            Fn[prev2_cell_id][i] * (xj - xjp1) * (xj - xjm1) / ((xjm2 - xjp1) * (xjm2 - xj) * (xjm2 - xjm1)) +
            Fn[prev_cell_id][i] * (xj - xjp1) * (xj - xjm2) / ((xjm1 - xjp1) * (xjm1 - xj) * (xjm1 - xjm2)) +
            Fn[cell_id][i] * (1. / (xj - xjm2) + 1. / (xj - xjm1) + 1. / (xj - xjp1)) +
            Fn[next_cell_id][i] * (xj - xjm1) * (xj - xjm2) / ((xjp1 - xj) * (xjp1 - xjm1) * (xjp1 - xjm2));
          // std::cout << "Aj = " << (xj - xjp1) * (xj - xjm1) / ((xjm2 - xjp1) * (xjm2 - xj) * (xjm2 - xjm1))
          //           << "; Bj = " << (xj - xjp1) * (xj - xjm2) / ((xjm1 - xjp1) * (xjm1 - xj) * (xjm1 - xjm2))
          //           << "; Cj = " << (1. / (xj - xjm2) + 1. / (xj - xjm1) + 1. / (xj - xjp1))
          //           << "; Dj = " << (xj - xjm1) * (xj - xjm2) / ((xjp1 - xj) * (xjp1 - xjm1) * (xjp1 - xjm2)) <<
          //           "\n";
        } else if (m_lambda_vector[i] < 0) {
          deltaFn[cell_id][i] =
            Fn[prev_cell_id][i] * (xj - xjp1) * (xj - xjp2) / ((xjm1 - xj) * (xjm1 - xjp1) * (xjm1 - xjp2)) +
            Fn[cell_id][i] * (1. / (xj - xjp2) + 1. / (xj - xjp1) + 1. / (xj - xjm1)) +
            Fn[next_cell_id][i] * (xj - xjm1) * (xj - xjp2) / ((xjp1 - xjm1) * (xjp1 - xj) * (xjp1 - xjp2)) +
            Fn[next2_cell_id][i] * (xj - xjm1) * (xj - xjp1) / ((xjp2 - xjm1) * (xjp2 - xj) * (xjp2 - xjp1));

        } else {
          deltaFn[cell_id][i] = 0;
        }
      }
    }
    return deltaFn;
  }

  DiscreteFunctionP0Vector<double>
  compute_delta2Fn_Unif(const DiscreteFunctionP0Vector<const double>& Fn)
  {
    const size_t nb_waves = m_lambda_vector.size();

    DiscreteFunctionP0Vector<double> deltaFn(p_mesh, nb_waves);

    for (CellId cell_id = 0; cell_id < m_mesh.numberOfCells(); ++cell_id) {
      for (size_t i_wave = 0; i_wave < nb_waves; ++i_wave) {
        if (m_lambda_vector[i_wave] < 0) {
          deltaFn[cell_id][i_wave] =
            -1. / 3 * Fn[(cell_id + m_mesh.numberOfCells() - 1) % m_mesh.numberOfCells()][i_wave] -
            0.5 * Fn[cell_id][i_wave] + Fn[(cell_id + 1) % m_mesh.numberOfCells()][i_wave] -
            1. / 6 * Fn[(cell_id + 2) % m_mesh.numberOfCells()][i_wave];
        } else if (m_lambda_vector[i_wave] > 0) {
          deltaFn[cell_id][i_wave] =
            1. / 3 * Fn[(cell_id + 1) % m_mesh.numberOfCells()][i_wave] + 0.5 * Fn[cell_id][i_wave] -
            Fn[(cell_id + m_mesh.numberOfCells() - 1) % m_mesh.numberOfCells()][i_wave] +
            1. / 6 * Fn[(cell_id + m_mesh.numberOfCells() - 2) % m_mesh.numberOfCells()][i_wave];
        } else {
          deltaFn[cell_id][i_wave] = 0;
        }
      }
    }
    return deltaFn;
  }

  DiscreteFunctionP0Vector<double>
  compute_delta4Fn_Unif(const DiscreteFunctionP0Vector<const double>& Fn)
  {
    const size_t nb_waves = m_lambda_vector.size();

    DiscreteFunctionP0Vector<double> deltaFn(p_mesh, nb_waves);

    for (CellId cell_id = 0; cell_id < m_mesh.numberOfCells(); ++cell_id) {
      for (size_t i_wave = 0; i_wave < nb_waves; ++i_wave) {
        // deltaFn[cell_id][i_wave] =
        //   -(1. / 12) * (Fn[(cell_id + 2) % m_mesh.numberOfCells()][i_wave] -
        //                 Fn[(cell_id + m_mesh.numberOfCells() - 2) % m_mesh.numberOfCells()][i_wave]) +
        //   (2. / 3) * (Fn[(cell_id + 1) % m_mesh.numberOfCells()][i_wave] -
        //               Fn[(cell_id + m_mesh.numberOfCells() - 1) % m_mesh.numberOfCells()][i_wave]);

        if (m_lambda_vector[i_wave] < 0) {
          deltaFn[cell_id][i_wave] = 1. / 12 * Fn[(cell_id + 3) % m_mesh.numberOfCells()][i_wave] -
                                     0.5 * Fn[(cell_id + 2) % m_mesh.numberOfCells()][i_wave] +
                                     3. / 2 * Fn[(cell_id + 1) % m_mesh.numberOfCells()][i_wave] -
                                     5. / 6 * Fn[cell_id][i_wave] -
                                     0.25 * Fn[(cell_id + m_mesh.numberOfCells() - 1) % m_mesh.numberOfCells()][i_wave];
        } else if (m_lambda_vector[i_wave] > 0) {
          deltaFn[cell_id][i_wave] =
            0.25 * Fn[(cell_id + 1) % m_mesh.numberOfCells()][i_wave] + 5. / 6 * Fn[cell_id][i_wave] -
            3. / 2 * Fn[(cell_id + m_mesh.numberOfCells() - 1) % m_mesh.numberOfCells()][i_wave] +
            0.5 * Fn[(cell_id + m_mesh.numberOfCells() - 2) % m_mesh.numberOfCells()][i_wave] -
            1. / 12 * Fn[(cell_id + m_mesh.numberOfCells() - 3) % m_mesh.numberOfCells()][i_wave];
        } else {
          deltaFn[cell_id][i_wave] = 0;
        }
      }
    }

    return deltaFn;
  }

  std::vector<CellValue<double>>
  compute_PFnp1(const DiscreteFunctionP0Vector<const double>& F0,
                std::vector<DiscreteFunctionP0Vector<double>>& deltaF,
                DiscreteFunctionP0Vector<double>& deltaF0,
                const TinyMatrix<2>& B,
                const TinyVector<2>& b0,   //  !! Impossible to put nb_subtimesteps for the dimension
                const double& dt)
  {
    const size_t nb_subtimesteps = deltaF.size();
    std::vector<CellValue<double>> PFnp1;
    for (size_t i_subtimestep = 0; i_subtimestep < nb_subtimesteps; ++i_subtimestep) {
      PFnp1.emplace_back(m_mesh.connectivity());
    }
    for (CellId cell_id = 0; cell_id < m_mesh.numberOfCells(); ++cell_id) {
      const double c = dt;
      for (size_t i_subtimestep = 0; i_subtimestep < nb_subtimesteps; ++i_subtimestep) {
        double PFn = 0;
        for (size_t i_wave = 0; i_wave < m_lambda_vector.size(); ++i_wave) {
          PFn += F0[cell_id][i_wave] -
                 c * m_lambda_vector[i_wave] *
                   (B(i_subtimestep, 0) * deltaF[0][cell_id][i_wave] +
                    B(i_subtimestep, 1) * deltaF[1][cell_id][i_wave] + b0[i_subtimestep] * deltaF0[cell_id][i_wave]);
        }

        PFnp1[i_subtimestep][cell_id] = PFn;
      }
    }
    return PFnp1;
  }

  CellValue<double>
  compute_PFn(const DiscreteFunctionP0Vector<const double> F)
  {
    CellValue<double> PFn{m_mesh.connectivity()};

    for (CellId cell_id = 0; cell_id < m_mesh.numberOfCells(); ++cell_id) {
      double PF = 0;
      for (size_t i = 0; i < m_lambda_vector.size(); ++i) {
        PF += F[cell_id][i];
      }
      PFn[cell_id] = PF;
    }
    return PFn;
  }

  std::tuple<std::shared_ptr<const DiscreteFunctionVariant>,
             std::shared_ptr<const DiscreteFunctionVariant>,
             std::shared_ptr<const DiscreteFunctionVariant>>
  apply(const double& eps)
  {
    // const auto& cell_to_node_matrix = m_mesh.connectivity().cellToNodeMatrix();
    const size_t nb_waves        = m_lambda_vector.size();
    const size_t nb_subtimesteps = 2;
    const size_t nb_equations    = 3;

    CellValue<bool> PFnp1_cell_is_valid{m_mesh.connectivity()};
    NodeValue<bool> PFnp1_node_is_valid{m_mesh.connectivity()};
    PFnp1_cell_is_valid.fill(true);
    PFnp1_node_is_valid.fill(true);

    std::vector<DiscreteFunctionP0Vector<double>> F_rho;
    std::vector<DiscreteFunctionP0Vector<double>> F_rho_u;
    std::vector<DiscreteFunctionP0Vector<double>> F_rho_E;

    std::vector<DiscreteFunctionP0Vector<double>> Fnp1_rho;
    std::vector<DiscreteFunctionP0Vector<double>> Fnp1_rho_u;
    std::vector<DiscreteFunctionP0Vector<double>> Fnp1_rho_E;

    for (size_t i_subtimestep = 0; i_subtimestep < nb_subtimesteps; ++i_subtimestep) {
      F_rho.emplace_back(p_mesh, nb_waves);
      F_rho_u.emplace_back(p_mesh, nb_waves);
      F_rho_E.emplace_back(p_mesh, nb_waves);

      Fnp1_rho.emplace_back(p_mesh, nb_waves);
      Fnp1_rho_u.emplace_back(p_mesh, nb_waves);
      Fnp1_rho_E.emplace_back(p_mesh, nb_waves);

      F_rho[i_subtimestep]   = copy(m_Fn_rho);
      F_rho_u[i_subtimestep] = copy(m_Fn_rho_u);
      F_rho_E[i_subtimestep] = copy(m_Fn_rho_E);

      Fnp1_rho[i_subtimestep]   = copy(m_Fn_rho);
      Fnp1_rho_u[i_subtimestep] = copy(m_Fn_rho_u);
      Fnp1_rho_E[i_subtimestep] = copy(m_Fn_rho_E);
    }

    DiscreteFunctionP0Vector<double> deltaFn_rho   = compute_delta3Fn(m_Fn_rho);
    DiscreteFunctionP0Vector<double> deltaFn_rho_u = compute_delta3Fn(m_Fn_rho_u);
    DiscreteFunctionP0Vector<double> deltaFn_rho_E = compute_delta3Fn(m_Fn_rho_E);

    const CellValue<const double> PFn_rho   = compute_PFn(m_Fn_rho);
    const CellValue<const double> PFn_rho_u = compute_PFn(m_Fn_rho_u);
    const CellValue<const double> PFn_rho_E = compute_PFn(m_Fn_rho_E);

    const CellArray<const double> APFn = getA(PFn_rho, PFn_rho_u, PFn_rho_E);
    const CellValue<double> APFn_rho{m_mesh.connectivity()};
    const CellValue<double> APFn_rho_u{m_mesh.connectivity()};
    const CellValue<double> APFn_rho_E{m_mesh.connectivity()};

    for (CellId cell_id = 0; cell_id < m_mesh.numberOfCells(); ++cell_id) {
      APFn_rho[cell_id]   = APFn[cell_id][0];
      APFn_rho_u[cell_id] = APFn[cell_id][1];
      APFn_rho_E[cell_id] = APFn[cell_id][2];
    }

    const CellArray<const double> MPFn_rho   = compute_M(PFn_rho, APFn_rho);
    const CellArray<const double> MPFn_rho_u = compute_M(PFn_rho_u, APFn_rho_u);
    const CellArray<const double> MPFn_rho_E = compute_M(PFn_rho_E, APFn_rho_E);

    for (size_t p = 0; p < 3; ++p) {
      const TinyMatrix<nb_subtimesteps> B(1. / 3., -1. / 24., 2. / 3., 1. / 6.);
      const TinyVector<nb_subtimesteps> b0(5. / 24., 1. / 6.);

      std::vector<DiscreteFunctionP0Vector<double>> deltaFnp1_rho;
      std::vector<DiscreteFunctionP0Vector<double>> deltaFnp1_rho_u;
      std::vector<DiscreteFunctionP0Vector<double>> deltaFnp1_rho_E;
      for (size_t i_subtimestep = 0; i_subtimestep < nb_subtimesteps; ++i_subtimestep) {
        deltaFnp1_rho.emplace_back(p_mesh, nb_waves);
        deltaFnp1_rho_u.emplace_back(p_mesh, nb_waves);
        deltaFnp1_rho_E.emplace_back(p_mesh, nb_waves);

        deltaFnp1_rho[i_subtimestep]   = compute_delta3Fn(Fnp1_rho[i_subtimestep]);
        deltaFnp1_rho_u[i_subtimestep] = compute_delta3Fn(Fnp1_rho_u[i_subtimestep]);
        deltaFnp1_rho_E[i_subtimestep] = compute_delta3Fn(Fnp1_rho_E[i_subtimestep]);
      }

      std::vector<CellValue<double>> PFnp1_rho;
      std::vector<CellValue<double>> PFnp1_rho_u;
      std::vector<CellValue<double>> PFnp1_rho_E;
      for (size_t i_subtimestep = 0; i_subtimestep < nb_subtimesteps; ++i_subtimestep) {
        PFnp1_rho.emplace_back(m_mesh.connectivity());
        PFnp1_rho_u.emplace_back(m_mesh.connectivity());
        PFnp1_rho_E.emplace_back(m_mesh.connectivity());
      }
      PFnp1_rho   = compute_PFnp1(m_Fn_rho, deltaFnp1_rho, deltaFn_rho, B, b0, m_dt);
      PFnp1_rho_u = compute_PFnp1(m_Fn_rho_u, deltaFnp1_rho_u, deltaFn_rho_u, B, b0, m_dt);
      PFnp1_rho_E = compute_PFnp1(m_Fn_rho_E, deltaFnp1_rho_E, deltaFn_rho_E, B, b0, m_dt);

      // std::cout << "PFnp1_rho_E(0) = " << PFnp1_rho_E[0] << "\n";

      std::vector<CellArray<double>> APFnp1;
      for (size_t i_subtimestep = 0; i_subtimestep < nb_subtimesteps; ++i_subtimestep) {
        APFnp1.emplace_back(m_mesh.connectivity(), nb_equations);

        APFnp1[i_subtimestep] = getA(PFnp1_rho[i_subtimestep], PFnp1_rho_u[i_subtimestep], PFnp1_rho_E[i_subtimestep]);
      }

      std::vector<CellValue<double>> APFnp1_rho;
      std::vector<CellValue<double>> APFnp1_rho_u;
      std::vector<CellValue<double>> APFnp1_rho_E;
      for (size_t i_subtimestep = 0; i_subtimestep < nb_subtimesteps; ++i_subtimestep) {
        APFnp1_rho.emplace_back(m_mesh.connectivity());
        APFnp1_rho_u.emplace_back(m_mesh.connectivity());
        APFnp1_rho_E.emplace_back(m_mesh.connectivity());
      }

      for (CellId cell_id = 0; cell_id < m_mesh.numberOfCells(); ++cell_id) {
        for (size_t i_subtimestep = 0; i_subtimestep < nb_subtimesteps; ++i_subtimestep) {
          APFnp1_rho[i_subtimestep][cell_id]   = APFnp1[i_subtimestep][cell_id][0];
          APFnp1_rho_u[i_subtimestep][cell_id] = APFnp1[i_subtimestep][cell_id][1];
          APFnp1_rho_E[i_subtimestep][cell_id] = APFnp1[i_subtimestep][cell_id][2];
        }
      }

      std::vector<CellArray<double>> MPFnp1_rho;
      std::vector<CellArray<double>> MPFnp1_rho_u;
      std::vector<CellArray<double>> MPFnp1_rho_E;

      for (size_t i_subtimestep = 0; i_subtimestep < nb_subtimesteps; ++i_subtimestep) {
        MPFnp1_rho.emplace_back(m_mesh.connectivity(), nb_waves);
        MPFnp1_rho_u.emplace_back(m_mesh.connectivity(), nb_waves);
        MPFnp1_rho_E.emplace_back(m_mesh.connectivity(), nb_waves);

        MPFnp1_rho[i_subtimestep]   = compute_M(PFnp1_rho[i_subtimestep], APFnp1_rho[i_subtimestep]);
        MPFnp1_rho_u[i_subtimestep] = compute_M(PFnp1_rho_u[i_subtimestep], APFnp1_rho_u[i_subtimestep]);
        MPFnp1_rho_E[i_subtimestep] = compute_M(PFnp1_rho_E[i_subtimestep], APFnp1_rho_E[i_subtimestep]);
      }

      const double c1 = 1. / (eps * eps + eps * m_dt * 0.5 + (1. / 12.) * m_dt * m_dt);
      const TinyMatrix<nb_subtimesteps> I_muB_inv(c1 * (eps + m_dt / 6.), c1 * m_dt / 24., c1 * (-m_dt * 2. / 3.),
                                                  c1 * (eps + m_dt / 3.));

      for (size_t i_subtimestep = 0; i_subtimestep < nb_subtimesteps; ++i_subtimestep) {
        for (CellId cell_id = 0; cell_id < m_mesh.numberOfCells(); ++cell_id) {
          const double c2 = (eps * m_dt);

          for (size_t i_wave = 0; i_wave < nb_waves; ++i_wave) {
            Fnp1_rho[i_subtimestep][cell_id][i_wave] =
              I_muB_inv(i_subtimestep, 0) *
                (m_dt * (B(0, 0) * MPFnp1_rho[0][cell_id][i_wave] + B(0, 1) * MPFnp1_rho[1][cell_id][i_wave]) +
                 eps * m_Fn_rho[cell_id][i_wave] -
                 c2 * m_lambda_vector[i_wave] *
                   (B(0, 0) * deltaFnp1_rho[0][cell_id][i_wave] + B(0, 1) * deltaFnp1_rho[1][cell_id][i_wave] +
                    b0[0] * deltaFn_rho[cell_id][i_wave]) +
                 m_dt * b0[0] * (MPFn_rho[cell_id][i_wave] - m_Fn_rho[cell_id][i_wave])) +
              I_muB_inv(i_subtimestep, 1) *
                (m_dt * (B(1, 0) * MPFnp1_rho[0][cell_id][i_wave] + B(1, 1) * MPFnp1_rho[1][cell_id][i_wave]) +
                 eps * m_Fn_rho[cell_id][i_wave] -
                 c2 * m_lambda_vector[i_wave] *
                   (B(1, 0) * deltaFnp1_rho[0][cell_id][i_wave] + B(1, 1) * deltaFnp1_rho[1][cell_id][i_wave] +
                    b0[1] * deltaFn_rho[cell_id][i_wave]) +
                 m_dt * b0[1] * (MPFn_rho[cell_id][i_wave] - m_Fn_rho[cell_id][i_wave]));

            Fnp1_rho_u[i_subtimestep][cell_id][i_wave] =
              I_muB_inv(i_subtimestep, 0) *
                (m_dt * (B(0, 0) * MPFnp1_rho_u[0][cell_id][i_wave] + B(0, 1) * MPFnp1_rho_u[1][cell_id][i_wave]) +
                 eps * m_Fn_rho_u[cell_id][i_wave] -
                 c2 * m_lambda_vector[i_wave] *
                   (B(0, 0) * deltaFnp1_rho_u[0][cell_id][i_wave] + B(0, 1) * deltaFnp1_rho_u[1][cell_id][i_wave] +
                    b0[0] * deltaFn_rho_u[cell_id][i_wave]) +
                 m_dt * b0[0] * (MPFn_rho_u[cell_id][i_wave] - m_Fn_rho_u[cell_id][i_wave])) +
              I_muB_inv(i_subtimestep, 1) *
                (m_dt * (B(1, 0) * MPFnp1_rho_u[0][cell_id][i_wave] + B(1, 1) * MPFnp1_rho_u[1][cell_id][i_wave]) +
                 eps * m_Fn_rho_u[cell_id][i_wave] -
                 c2 * m_lambda_vector[i_wave] *
                   (B(1, 0) * deltaFnp1_rho_u[0][cell_id][i_wave] + B(1, 1) * deltaFnp1_rho_u[1][cell_id][i_wave] +
                    b0[1] * deltaFn_rho_u[cell_id][i_wave]) +
                 m_dt * b0[1] * (MPFn_rho_u[cell_id][i_wave] - m_Fn_rho_u[cell_id][i_wave]));

            Fnp1_rho_E[i_subtimestep][cell_id][i_wave] =
              I_muB_inv(i_subtimestep, 0) *
                (m_dt * (B(0, 0) * MPFnp1_rho_E[0][cell_id][i_wave] + B(0, 1) * MPFnp1_rho_E[1][cell_id][i_wave]) +
                 eps * m_Fn_rho_E[cell_id][i_wave] -
                 c2 * m_lambda_vector[i_wave] *
                   (B(0, 0) * deltaFnp1_rho_E[0][cell_id][i_wave] + B(0, 1) * deltaFnp1_rho_E[1][cell_id][i_wave] +
                    b0[0] * deltaFn_rho_E[cell_id][i_wave]) +
                 m_dt * b0[0] * (MPFn_rho_E[cell_id][i_wave] - m_Fn_rho_E[cell_id][i_wave])) +
              I_muB_inv(i_subtimestep, 1) *
                (m_dt * (B(1, 0) * MPFnp1_rho_E[0][cell_id][i_wave] + B(1, 1) * MPFnp1_rho_E[1][cell_id][i_wave]) +
                 eps * m_Fn_rho_E[cell_id][i_wave] -
                 c2 * m_lambda_vector[i_wave] *
                   (B(1, 0) * deltaFnp1_rho_E[0][cell_id][i_wave] + B(1, 1) * deltaFnp1_rho_E[1][cell_id][i_wave] +
                    b0[1] * deltaFn_rho_E[cell_id][i_wave]) +
                 m_dt * b0[1] * (MPFn_rho_E[cell_id][i_wave] - m_Fn_rho_E[cell_id][i_wave]));

            // std::cout << "iteration p = " << p << "; cell_id = " << cell_id << "; i_subtimestep = " << i_subtimestep
            //           << "; i_wave = " << i_wave << "; F = " << Fnp1_rho[i_subtimestep][cell_id][i_wave] << ", "
            //           << Fnp1_rho_u[i_subtimestep][cell_id][i_wave] << ", "
            //           << Fnp1_rho_E[i_subtimestep][cell_id][i_wave] << "\n";
          }
        }

        PFnp1_node_is_valid.fill(true);
        F_rho[i_subtimestep]   = copy(Fnp1_rho[i_subtimestep]);
        F_rho_u[i_subtimestep] = copy(Fnp1_rho_u[i_subtimestep]);
        F_rho_E[i_subtimestep] = copy(Fnp1_rho_E[i_subtimestep]);
      }
    }

    return std::make_tuple(std::make_shared<DiscreteFunctionVariant>(Fnp1_rho[nb_subtimesteps - 1]),
                           std::make_shared<DiscreteFunctionVariant>(Fnp1_rho_u[nb_subtimesteps - 1]),
                           std::make_shared<DiscreteFunctionVariant>(Fnp1_rho_E[nb_subtimesteps - 1]));
  }

  EulerKineticSolverThirdOrderFD(std::shared_ptr<const MeshType> mesh,
                                 const double& dt,
                                 const double& gamma,
                                 const std::vector<double>& lambda_vector,
                                 const DiscreteFunctionP0Vector<const double>& Fn_rho,
                                 const DiscreteFunctionP0Vector<const double>& Fn_rho_u,
                                 const DiscreteFunctionP0Vector<const double>& Fn_rho_E)
    : m_dt{dt},
      m_lambda_vector{lambda_vector},
      m_Fn_rho{Fn_rho},
      m_Fn_rho_u{Fn_rho_u},
      m_Fn_rho_E{Fn_rho_E},
      m_gamma{gamma},
      p_mesh{mesh},
      m_mesh{*mesh}
  {
    if ((lambda_vector.size() != Fn_rho.size()) or
        (lambda_vector.size() != Fn_rho_u.size() or (lambda_vector.size() != Fn_rho_u.size()))) {
      throw NormalError("Incompatible dimensions of lambda vector and Fn");
    }
  }

  ~EulerKineticSolverThirdOrderFD() = default;
};

std::tuple<std::shared_ptr<const DiscreteFunctionVariant>,
           std::shared_ptr<const DiscreteFunctionVariant>,
           std::shared_ptr<const DiscreteFunctionVariant>>
eulerKineticSolverThirdOrderFD(const double& dt,
                               const double& gamma,
                               const std::vector<double>& lambda_vector,
                               const double& eps,
                               const std::shared_ptr<const DiscreteFunctionVariant>& F_rho,
                               const std::shared_ptr<const DiscreteFunctionVariant>& F_rho_u,
                               const std::shared_ptr<const DiscreteFunctionVariant>& F_rho_E)
{
  const std::shared_ptr<const MeshVariant> mesh_v = getCommonMesh({F_rho, F_rho_u, F_rho_E});
  if (mesh_v.use_count() == 0) {
    throw NormalError("F_rho, F_rho_u and F_rho_E have to be defined on the same mesh");
  }

  return std::visit(
    [&](auto&& p_mesh)
      -> std::tuple<std::shared_ptr<const DiscreteFunctionVariant>, std::shared_ptr<const DiscreteFunctionVariant>,
                    std::shared_ptr<const DiscreteFunctionVariant>> {
      using MeshType = mesh_type_t<decltype(p_mesh)>;
      if constexpr (is_polygonal_mesh_v<MeshType> and (MeshType::Dimension == 1)) {
        auto Fn_rho   = F_rho->get<DiscreteFunctionP0Vector<const double>>();
        auto Fn_rho_u = F_rho_u->get<DiscreteFunctionP0Vector<const double>>();
        auto Fn_rho_E = F_rho_E->get<DiscreteFunctionP0Vector<const double>>();

        EulerKineticSolverThirdOrderFD solver(p_mesh, dt, gamma, lambda_vector, Fn_rho, Fn_rho_u, Fn_rho_E);

        return solver.apply(eps);
      } else {
        throw NotImplementedError("Invalid mesh type for Scalar Kinetic solver");
      }
    },
    mesh_v->variant());
}
